@@ -5,25 +5,31 @@ equations using matrix operations.
 
 from copy import deepcopy
 from itertools import combinations
-from typing import List
+from math import comb
+from typing import Union
 
 import numpy as np
+import ray
 from numba import njit, prange
-from pymatgen.core.composition import Composition
-from scipy.special import comb
+from pymatgen.core.composition import Element
 from tqdm import tqdm
 
-from rxn_network.core import CostFunction, Pathway, Solver
+from rxn_network.core.composition import Composition
+from rxn_network.core.cost_function import CostFunction
+from rxn_network.core.solver import Solver
 from rxn_network.entries.entry_set import GibbsEntrySet
 from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
-
 from rxn_network.enumerators.minimize import (
     MinimizeGibbsEnumerator,
     MinimizeGrandPotentialEnumerator,
 )
 from rxn_network.pathways.balanced import BalancedPathway
+from rxn_network.pathways.pathway_set import PathwaySet
 from rxn_network.reactions.computed import ComputedReaction
-from rxn_network.utils import grouper
+from rxn_network.reactions.open import OpenComputedReaction
+from rxn_network.reactions.reaction_set import ReactionSet
+from rxn_network.utils.funcs import grouper
+from rxn_network.utils.ray import initialize_ray, to_iterator
 
 
 class PathwaySolver(Solver):
@@ -33,54 +39,58 @@ class PathwaySolver(Solver):
     lists of reactions)
     """
 
-    BATCH_SIZE = 500000  # how many reactions to process in parallel using numba
-
     def __init__(
         self,
+        pathways: PathwaySet,
         entries: GibbsEntrySet,
-        pathways: List[Pathway],
         cost_function: CostFunction,
         open_elem: str = None,
         chempot: float = None,
+        chunk_size=100000,
+        batch_size=None,
     ):
         """
         Args:
-            entries: GibbsEntrySet containing all entries in the network.
             pathways: List of reaction pathways derived from the network.
+            entries: GibbsEntrySet containing all entries in the network.
             cost_function: CostFunction object to use for the solver.
             open_elem: Element to use for pathways with an open element.
             chempot: Chemical potential to use for pathways with an open element.
         """
-        super().__init__(entries=deepcopy(entries), pathways=deepcopy(pathways))
+        super().__init__(pathways=deepcopy(pathways))
+        self._entries = entries
         self.cost_function = cost_function
-        self.open_elem = open_elem
+        self.open_elem = Element(open_elem) if open_elem else None
         self.chempot = chempot
+        self.chunk_size = chunk_size
+        self.batch_size = batch_size
 
     def solve(
         self,
-        net_rxn: ComputedReaction,
+        net_rxn: Union[ComputedReaction, OpenComputedReaction],
         max_num_combos: int = 4,
         find_intermediate_rxns: bool = True,
         intermediate_rxn_energy_cutoff: float = 0.0,
         use_basic_enumerator: bool = True,
         use_minimize_enumerator: bool = False,
         filter_interdependent: bool = True,
-    ) -> List[BalancedPathway]:
+    ) -> PathwaySet:
         """
 
         Args:
             net_rxn: The reaction representing the total reaction from precursors to
                 final targets.
             max_num_combos: The maximum allowable size of the balanced reaction pathway.
-                At 5 or more, the solver will be very slow.
+                At values <=5, the solver will start to take a significant amount of
+                time to run.
             find_intermediate_rxns: Whether to find intermediate reactions; crucial for
-                finding pathways where intermediates react together, as these reactions may
-                not occur in the graph-derived pathways. Defaults to True.
+                finding pathways where intermediates react together, as these reactions
+                may not occur in the graph-derived pathways. Defaults to True.
             intermediate_rxn_energy_cutoff: An energy cutoff by which to filter down
-                intermediate reactions. This can be useful when there are a large number of
-                possible intermediates. < 0 means allow only exergonic reactions.
-            use_basic_enumerator: Whether to use the BasicEnumerator to find intermediate
-                reactions. Defaults to True.
+                intermediate reactions. This can be useful when there are a large number
+                of possible intermediates. < 0 means allow only exergonic reactions.
+            use_basic_enumerator: Whether to use the BasicEnumerator to find
+                intermediate reactions. Defaults to True.
             use_minimize_enumerator: Whether to use the MinimizeGibbsEnumerator to find
                 intermediate reactions. Defaults to False.
             filter_interdependent: Whether or not to filter out pathways where reaction
@@ -89,8 +99,16 @@ class PathwaySolver(Solver):
         Returns:
             A list of BalancedPathway objects.
         """
-        entries = deepcopy(self.entries)
-        entries = entries.entries_list
+
+        if not net_rxn.balanced:
+            raise ValueError(
+                "Net reaction must be balanceable to find all reaction pathways."
+            )
+
+        initialize_ray()
+
+        entries_copy = deepcopy(self.entries)
+        entries = entries_copy.entries_list
         num_entries = len(entries)
 
         reactions = deepcopy(self.reactions)
@@ -99,12 +117,7 @@ class PathwaySolver(Solver):
         precursors = deepcopy(net_rxn.reactant_entries)
         targets = deepcopy(net_rxn.product_entries)
 
-        if not net_rxn.balanced:
-            raise ValueError(
-                "Net reaction must be balanceable to find all reaction pathways."
-            )
-
-        self.logger.info(f"NET RXN: {net_rxn} \n")
+        self.logger.info(f"Net reaction: {net_rxn} \n")
 
         if find_intermediate_rxns:
             self.logger.info("Identifying reactions between intermediates...")
@@ -116,74 +129,70 @@ class PathwaySolver(Solver):
                 use_minimize_enumerator,
             )
             intermediate_costs = [
-                self.cost_function.evaluate(r) for r in intermediate_rxns
+                self.cost_function.evaluate(r) for r in intermediate_rxns.get_rxns()
             ]
             for r, c in zip(intermediate_rxns, intermediate_costs):
                 if r not in reactions:
                     reactions.append(r)
                     costs.append(c)
 
-        net_rxn_vector = self._build_idx_vector(net_rxn, num_entries)
+        net_rxn_vector = net_rxn.get_entry_idx_vector(num_entries)
+
         if net_rxn in reactions:
             reactions.remove(net_rxn)
 
+        reaction_set = ray.put(ReactionSet.from_rxns(reactions))
+        entries = ray.put(entries)
+        costs = ray.put(costs)
+        num_entries = ray.put(num_entries)
+        net_rxn_vector = ray.put(net_rxn_vector)
+        open_elem = ray.put(self.open_elem)
+        chempot = ray.put(self.chempot)
+
+        num_rxns = len(reactions)
+        batch_size = self.batch_size or ray.cluster_resources()["CPU"] - 1
+
+        num_combos = sum(comb(num_rxns, k) for k in range(1, max_num_combos + 1))
+        num_batches = int((num_combos // self.chunk_size + 1) // batch_size + 1)
+
         paths = []
+        paths_refs = []
+        batch_count = 1
         for n in range(1, max_num_combos + 1):
-            total = int(comb(len(reactions), n) / self.BATCH_SIZE) + 1
-            groups = grouper(combinations(range(len(reactions)), n), self.BATCH_SIZE)
-
-            pbar = groups
-            if n >= 4:
-                self.logger.info(f"Solving for balanced pathways of size {n} ...")
-                pbar = tqdm(groups, total=total, desc="PathwaySolver")
-
-            all_c_mats, all_m_mats = [], []
-            for idx, combos in enumerate(pbar):
-                if n >= 4:
-                    pbar.set_description(
-                        f"{self.BATCH_SIZE*idx}/{total*self.BATCH_SIZE}"
+            for group in grouper(combinations(range(num_rxns), n), self.chunk_size):
+                paths_refs.append(
+                    _get_balanced_paths_ray.remote(
+                        group,
+                        reaction_set,
+                        costs,
+                        entries,
+                        num_entries,
+                        net_rxn_vector,
+                        open_elem,
+                        chempot,
                     )
-                comp_matrices = np.stack(
-                    [
-                        np.vstack(
-                            [
-                                self._build_idx_vector(reactions[r], num_entries)
-                                for r in combo
-                            ]
-                        )
-                        for combo in combos
-                        if combo
-                    ]
                 )
-                c_mats, m_mats = balance_path_arrays(comp_matrices, net_rxn_vector)
-                all_c_mats.extend(c_mats)
-                all_m_mats.extend(m_mats)
+                if len(paths_refs) >= batch_size:
+                    for paths_ref in tqdm(
+                        to_iterator(paths_refs),
+                        total=len(paths_refs),
+                        desc=(
+                            f"{self.__class__.__name__} (Batch"
+                            f" {batch_count}/{num_batches})"
+                        ),
+                    ):
+                        paths.extend(paths_ref)
 
-            for c_mat, m_mat in zip(all_c_mats, all_m_mats):
-                path_rxns = []
-                path_costs = []
-                for rxn_mat in c_mat:
-                    ents, coeffs = zip(
-                        *[
-                            (entries[idx], c)
-                            for idx, c in enumerate(rxn_mat)
-                            if not np.isclose(c, 0.0)
-                        ]
-                    )
+                    batch_count += 1
 
-                    rxn = ComputedReaction(entries=ents, coefficients=coeffs)
+                    paths_refs = []
 
-                    try:
-                        path_rxns.append(rxn)
-                        path_costs.append(costs[reactions.index(rxn)])
-                    except Exception as e:
-                        print(e)
-                        continue
-
-                p = BalancedPathway(
-                    path_rxns, m_mat.flatten(), path_costs, balanced=True
-                )
-                paths.append(p)
+        for paths_ref in tqdm(
+            to_iterator(paths_refs),
+            total=len(paths_refs),
+            desc=f"{self.__class__.__name__} (Batch {batch_count}/{num_batches})",
+        ):
+            paths.extend(paths_ref)
 
         filtered_paths = []
         if filter_interdependent:
@@ -196,32 +205,8 @@ class PathwaySolver(Solver):
             filtered_paths = paths
 
         filtered_paths = sorted(list(set(filtered_paths)), key=lambda p: p.average_cost)
-        return filtered_paths
 
-    @staticmethod
-    def _build_idx_vector(rxn: ComputedReaction, num_entries: int) -> np.ndarray:
-        """
-        Builds a vector of indices for a reaction based on the data["idx"] attribute of
-        each entry in the reaction. This allows for reactions to be more easily
-        represented as vectors.
-
-        Args:
-            rxn: a ComputedReaction object
-            num_entries: The number of total entries in the entry set (that the reaction
-                comes from).
-
-        Returns:
-            A vector of indices for the reaction.
-        """
-        indices = [e.data.get("idx") for e in rxn.entries]
-        if None in indices:
-            raise ValueError(
-                f"Could not find index for one or more entries in reaction: {rxn}"
-            )
-
-        v = np.zeros(num_entries)
-        v[indices] = rxn.coefficients
-        return v
+        return PathwaySet.from_paths(filtered_paths)
 
     def _find_intermediate_rxns(
         self,
@@ -231,10 +216,9 @@ class PathwaySolver(Solver):
         use_minimize_enumerator,
     ):
         """
-        Internal method for finding intermediate reactions using enumerators and
+        Method for finding intermediate reactions using enumerators and
         specified settings.
         """
-        rxns = []
 
         intermediates = {e for rxn in self.reactions for e in rxn.entries}
         intermediates = GibbsEntrySet(
@@ -243,9 +227,20 @@ class PathwaySolver(Solver):
         target_formulas = [e.composition.reduced_formula for e in targets]
         ref_elems = {e for e in self.entries if e.is_element}
 
+        intermediates = intermediates | ref_elems
+
+        rxn_set = ReactionSet(
+            intermediates.entries_list,
+            [],
+            [],
+            open_elem=self.open_elem,
+            chempot=self.chempot,
+            all_data=[],
+        )
+
         if use_basic_enumerator:
             be = BasicEnumerator(targets=target_formulas, calculate_e_above_hulls=False)
-            rxns.extend(be.enumerate(intermediates))
+            rxn_set = rxn_set.add_rxn_set(be.enumerate(intermediates))
 
             if self.open_elem:
                 boe = BasicOpenEnumerator(
@@ -254,37 +249,39 @@ class PathwaySolver(Solver):
                     calculate_e_above_hulls=False,
                 )
 
-                rxns.extend(boe.enumerate(intermediates))
+                rxn_set = rxn_set.add_rxn_set(boe.enumerate(intermediates))
 
         if use_minimize_enumerator:
-            ents = deepcopy(intermediates)
-            ents = ents | ref_elems
-
             mge = MinimizeGibbsEnumerator(
                 targets=target_formulas, calculate_e_above_hulls=False
             )
-            rxns.extend(mge.enumerate(ents))
+            rxn_set = rxn_set.add_rxn_set(mge.enumerate(intermediates))
 
             if self.open_elem:
                 mgpe = MinimizeGrandPotentialEnumerator(
                     open_elem=self.open_elem,
                     mu=self.chempot,
                     targets=target_formulas,
-                    calculate_e_above_hulls=False,
                 )
-                rxns.extend(mgpe.enumerate(ents))
+                rxn_set.add_rxn_set(mgpe.enumerate(intermediates))
 
-        rxns = list(filter(lambda x: x.energy_per_atom < energy_cutoff, rxns))
+        rxns = list(filter(lambda x: x.energy_per_atom < energy_cutoff, rxn_set))
         rxns = [r for r in rxns if all(e in intermediates for e in r.entries)]
-        rxns = [r for r in rxns if (len(r.reactants) < 4 and len(r.products) < 4)]
+        num_rxns = len(rxns)
+        rxns = ReactionSet.from_rxns(rxns, filter_duplicates=True)
 
-        self.logger.info(f"Found {len(rxns)} intermediate reactions!")
+        self.logger.info(f"Found {num_rxns} intermediate reactions! \n")
 
         return rxns
 
+    @property
+    def entries(self) -> GibbsEntrySet:
+        """Entry set used in solver"""
+        return self._entries
 
-@njit(parallel=True)
-def balance_path_arrays(
+
+@njit(parallel=True, fastmath=True)
+def _balance_path_arrays(
     comp_matrices: np.ndarray,
     net_coeffs: np.ndarray,
     tol: float = 1e-6,
@@ -343,3 +340,69 @@ def balance_path_arrays(
         filtered_multiplicities[i] = all_multiplicities[idx]
 
     return filtered_comp_matrices, filtered_multiplicities
+
+
+def _create_comp_matrices(combos, rxns, num_entries):
+    """Create array of stoichiometric coefficients for each reaction."""
+    comp_matrices = np.stack(
+        [
+            np.vstack([rxns[r].get_entry_idx_vector(num_entries) for r in combo])
+            for combo in combos
+            if combo
+        ]
+    )
+    return comp_matrices
+
+
+@ray.remote
+def _get_balanced_paths_ray(
+    combos,
+    reaction_set,
+    costs,
+    entries,
+    num_entries,
+    net_rxn_vector,
+    open_elem,
+    chempot,
+):
+    reactions = list(reaction_set.get_rxns())
+    comp_matrices = _create_comp_matrices(combos, reactions, num_entries)
+
+    paths = []
+
+    c_mats, m_mats = _balance_path_arrays(comp_matrices, net_rxn_vector)
+
+    for c_mat, m_mat in zip(c_mats, m_mats):
+        path_rxns = []
+        path_costs = []
+
+        for rxn_mat in c_mat:
+            ents, coeffs = zip(
+                *[
+                    (entries[idx], c)
+                    for idx, c in enumerate(rxn_mat)
+                    if not np.isclose(c, 0.0)
+                ]
+            )
+
+            if open_elem is not None:
+                rxn = OpenComputedReaction(
+                    entries=ents,
+                    coefficients=coeffs,
+                    chempots={open_elem: chempot},
+                )
+
+            else:
+                rxn = ComputedReaction(entries=ents, coefficients=coeffs)
+
+            try:
+                path_rxns.append(rxn)
+                path_costs.append(costs[reactions.index(rxn)])
+            except Exception as e:
+                print(e)
+                continue
+
+        p = BalancedPathway(path_rxns, m_mat.flatten(), path_costs, balanced=True)
+        paths.append(p)
+
+    return paths
